# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from orquesta import events
from orquesta import exceptions as exc
from orquesta import statuses
from orquesta.utils import jsonify as json_util


LOG = logging.getLogger(__name__)


WORKFLOW_STATE_MACHINE_DATA = {
    statuses.UNSET: {
        events.WORKFLOW_REQUESTED: statuses.REQUESTED,
        events.WORKFLOW_SCHEDULED: statuses.SCHEDULED,
        events.WORKFLOW_DELAYED: statuses.DELAYED,
        events.WORKFLOW_RUNNING: statuses.RUNNING,
        events.WORKFLOW_FAILED: statuses.FAILED,
    },
    statuses.REQUESTED: {
        events.WORKFLOW_SCHEDULED: statuses.SCHEDULED,
        events.WORKFLOW_DELAYED: statuses.DELAYED,
        events.WORKFLOW_RUNNING: statuses.RUNNING,
        events.WORKFLOW_PAUSING_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_PAUSED_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_CANCELING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_FAILED: statuses.FAILED,
    },
    statuses.SCHEDULED: {
        events.WORKFLOW_DELAYED: statuses.DELAYED,
        events.WORKFLOW_RUNNING: statuses.RUNNING,
        events.WORKFLOW_PAUSING_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_PAUSED_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_CANCELING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_FAILED: statuses.FAILED,
    },
    statuses.DELAYED: {
        events.WORKFLOW_SCHEDULED: statuses.SCHEDULED,
        events.WORKFLOW_RUNNING: statuses.RUNNING,
        events.WORKFLOW_PAUSING_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_PAUSED_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_CANCELING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_FAILED: statuses.FAILED,
    },
    statuses.RUNNING: {
        events.WORKFLOW_PAUSING_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.WORKFLOW_PAUSING_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_PAUSED_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.WORKFLOW_PAUSED_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_CANCELING_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.WORKFLOW_CANCELING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_CANCELED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.WORKFLOW_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_FAILED: statuses.FAILED,
        events.WORKFLOW_SUCCEEDED: statuses.SUCCEEDED,
        events.TASK_PENDING_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.TASK_PENDING_WORKFLOW_DORMANT: statuses.PAUSED,
        # Workflow status is not affected by task going into pause statuses. This allows for more
        # granular control if operators want to pause a task but keep the rest of running.
        events.TASK_PAUSING: statuses.RUNNING,
        events.TASK_PAUSED_WORKFLOW_ACTIVE: statuses.RUNNING,
        # But if the workflow is not active, then pause the workflow when the last task paused.
        events.TASK_PAUSED_WORKFLOW_DORMANT: statuses.PAUSED,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_INCOMPLETE: statuses.RUNNING,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_COMPLETED: statuses.RUNNING,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_PAUSED: statuses.PAUSING,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_CANCELED: statuses.CANCELING,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_INCOMPLETE: statuses.RUNNING,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_COMPLETED: statuses.SUCCEEDED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_PAUSED: statuses.PAUSED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_CANCELED: statuses.CANCELED,
        # Fail the workflow quickly even if there are still running tasks.
        events.TASK_FAILED_WORKFLOW_ACTIVE: statuses.FAILED,
        events.TASK_FAILED_WORKFLOW_DORMANT: statuses.FAILED,
        # A task is remediated if it failed but there is one or more task transitions.
        # Therefore, remediated tasks are treated as if they are succeeded tasks.
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_INCOMPLETE: statuses.RUNNING,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_COMPLETED: statuses.RUNNING,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_PAUSED: statuses.PAUSING,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_CANCELED: statuses.CANCELING,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_INCOMPLETE: statuses.RUNNING,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_COMPLETED: statuses.SUCCEEDED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_PAUSED: statuses.PAUSED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_CANCELED: statuses.CANCELED,
        events.TASK_CANCELING: statuses.CANCELING,
        events.TASK_CANCELED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.TASK_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
    },
    statuses.PAUSING: {
        events.WORKFLOW_PAUSING_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.WORKFLOW_PAUSING_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_PAUSED_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.WORKFLOW_PAUSED_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_CANCELING_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.WORKFLOW_CANCELING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_CANCELED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.WORKFLOW_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_RUNNING: statuses.RUNNING,
        events.WORKFLOW_RESUMING: statuses.RESUMING,
        events.WORKFLOW_FAILED: statuses.FAILED,
        events.TASK_PENDING_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.TASK_PENDING_WORKFLOW_DORMANT: statuses.PAUSED,
        events.TASK_PAUSED_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.TASK_PAUSED_WORKFLOW_DORMANT: statuses.PAUSED,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_INCOMPLETE: statuses.PAUSING,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_COMPLETED: statuses.PAUSING,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_PAUSED: statuses.PAUSING,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_CANCELED: statuses.PAUSING,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_INCOMPLETE: statuses.PAUSED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_COMPLETED: statuses.PAUSED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_PAUSED: statuses.PAUSED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_CANCELED: statuses.CANCELED,
        events.TASK_FAILED_WORKFLOW_ACTIVE: statuses.FAILED,
        events.TASK_FAILED_WORKFLOW_DORMANT: statuses.FAILED,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_INCOMPLETE: statuses.PAUSING,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_COMPLETED: statuses.PAUSING,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_PAUSED: statuses.PAUSING,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_CANCELED: statuses.CANCELING,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_INCOMPLETE: statuses.PAUSED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_COMPLETED: statuses.PAUSED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_PAUSED: statuses.PAUSED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_CANCELED: statuses.CANCELED,
        events.TASK_CANCELING: statuses.CANCELING,
        events.TASK_CANCELED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.TASK_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.TASK_RETRYING_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.TASK_RETRYING_WORKFLOW_DORMANT: statuses.PAUSED,
    },
    statuses.PAUSED: {
        events.WORKFLOW_RUNNING: statuses.RUNNING,
        events.WORKFLOW_RUNNING_WORKFLOW_COMPLETED: statuses.SUCCEEDED,
        events.WORKFLOW_RESUMING: statuses.RESUMING,
        events.WORKFLOW_RESUMING_WORKFLOW_COMPLETED: statuses.SUCCEEDED,
        events.WORKFLOW_CANCELING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_FAILED: statuses.FAILED,
        events.TASK_RUNNING: statuses.RUNNING,
        events.TASK_RESUMING: statuses.RUNNING,
    },
    statuses.RESUMING: {
        events.WORKFLOW_PAUSING_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.WORKFLOW_PAUSING_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_PAUSED_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.WORKFLOW_PAUSED_WORKFLOW_DORMANT: statuses.PAUSED,
        events.WORKFLOW_CANCELING_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.WORKFLOW_CANCELING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_CANCELED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.WORKFLOW_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_RUNNING: statuses.RUNNING,
        events.WORKFLOW_FAILED: statuses.FAILED,
        events.TASK_RUNNING: statuses.RUNNING,
        events.TASK_PENDING_WORKFLOW_ACTIVE: statuses.PAUSING,
        events.TASK_PENDING_WORKFLOW_DORMANT: statuses.PAUSED,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_INCOMPLETE: statuses.RUNNING,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_COMPLETED: statuses.RUNNING,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_INCOMPLETE: statuses.RUNNING,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_COMPLETED: statuses.SUCCEEDED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_PAUSED: statuses.PAUSED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_CANCELED: statuses.CANCELED,
        events.TASK_FAILED_WORKFLOW_ACTIVE: statuses.FAILED,
        events.TASK_FAILED_WORKFLOW_DORMANT: statuses.FAILED,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_INCOMPLETE: statuses.RUNNING,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_COMPLETED: statuses.RUNNING,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_INCOMPLETE: statuses.RUNNING,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_COMPLETED: statuses.SUCCEEDED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_PAUSED: statuses.PAUSED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_CANCELED: statuses.CANCELED,
        events.TASK_CANCELING: statuses.CANCELING,
        events.TASK_CANCELED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.TASK_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
    },
    statuses.CANCELING: {
        events.WORKFLOW_CANCELING_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.WORKFLOW_CANCELING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_CANCELED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.WORKFLOW_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.WORKFLOW_FAILED: statuses.FAILED,
        events.TASK_PENDING_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.TASK_PENDING_WORKFLOW_DORMANT: statuses.CANCELED,
        events.TASK_PAUSED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.TASK_PAUSED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_INCOMPLETE: statuses.CANCELING,
        events.TASK_SUCCEEDED_WORKFLOW_ACTIVE_COMPLETED: statuses.CANCELING,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_INCOMPLETE: statuses.CANCELED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_COMPLETED: statuses.CANCELED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_PAUSED: statuses.CANCELED,
        events.TASK_SUCCEEDED_WORKFLOW_DORMANT_CANCELED: statuses.CANCELED,
        events.TASK_FAILED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.TASK_FAILED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_INCOMPLETE: statuses.CANCELING,
        events.TASK_REMEDIATED_WORKFLOW_ACTIVE_COMPLETED: statuses.CANCELING,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_INCOMPLETE: statuses.CANCELED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_COMPLETED: statuses.CANCELED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_PAUSED: statuses.CANCELED,
        events.TASK_REMEDIATED_WORKFLOW_DORMANT_CANCELED: statuses.CANCELED,
        events.TASK_CANCELED_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.TASK_CANCELED_WORKFLOW_DORMANT: statuses.CANCELED,
        events.TASK_RETRYING_WORKFLOW_ACTIVE: statuses.CANCELING,
        events.TASK_RETRYING_WORKFLOW_DORMANT: statuses.CANCELED,
    },
    statuses.CANCELED: {},
    statuses.SUCCEEDED: {
        # Workflow status can transition from succeeded to failed in  cases
        # where there is exception while rendering workflow output.
        events.WORKFLOW_FAILED: statuses.FAILED
    },
    statuses.FAILED: {},
}


TASK_STATE_MACHINE_DATA = {
    statuses.UNSET: {
        events.ACTION_REQUESTED: statuses.REQUESTED,
        events.ACTION_SCHEDULED: statuses.SCHEDULED,
        events.ACTION_DELAYED: statuses.DELAYED,
        events.ACTION_RUNNING: statuses.RUNNING,
        events.ACTION_PENDING: statuses.PENDING,
        events.TASK_CONTINUE_REQUESTED: statuses.SUCCEEDED,
        events.TASK_NOOP_REQUESTED: statuses.SUCCEEDED,
        events.TASK_FAIL_REQUESTED: statuses.FAILED,
    },
    statuses.REQUESTED: {
        events.ACTION_SCHEDULED: statuses.SCHEDULED,
        events.ACTION_DELAYED: statuses.DELAYED,
        events.ACTION_RUNNING: statuses.RUNNING,
        events.ACTION_PENDING: statuses.PENDING,
        events.ACTION_PAUSING: statuses.PAUSING,
        events.ACTION_PAUSED: statuses.PAUSED,
        events.ACTION_CANCELING: statuses.CANCELING,
        events.ACTION_CANCELED: statuses.CANCELED,
        events.ACTION_FAILED: statuses.FAILED,
        events.ACTION_EXPIRED: statuses.EXPIRED,
        events.ACTION_ABANDONED: statuses.ABANDONED,
    },
    statuses.SCHEDULED: {
        events.ACTION_DELAYED: statuses.DELAYED,
        events.ACTION_RUNNING: statuses.RUNNING,
        events.ACTION_PENDING: statuses.PENDING,
        events.ACTION_PAUSING: statuses.PAUSING,
        events.ACTION_PAUSED: statuses.PAUSED,
        events.ACTION_CANCELING: statuses.CANCELING,
        events.ACTION_CANCELED: statuses.CANCELED,
        events.ACTION_FAILED: statuses.FAILED,
        events.ACTION_EXPIRED: statuses.EXPIRED,
        events.ACTION_ABANDONED: statuses.ABANDONED,
    },
    statuses.DELAYED: {
        events.ACTION_SCHEDULED: statuses.SCHEDULED,
        events.ACTION_RUNNING: statuses.RUNNING,
        events.ACTION_PENDING: statuses.PENDING,
        events.ACTION_PAUSING: statuses.PAUSING,
        events.ACTION_PAUSED: statuses.PAUSED,
        events.ACTION_CANCELING: statuses.CANCELING,
        events.ACTION_CANCELED: statuses.CANCELED,
        events.ACTION_FAILED: statuses.FAILED,
        events.ACTION_EXPIRED: statuses.EXPIRED,
        events.ACTION_ABANDONED: statuses.ABANDONED,
    },
    statuses.RUNNING: {
        events.ACTION_RUNNING: statuses.RUNNING,
        events.ACTION_PENDING: statuses.PENDING,
        events.ACTION_PENDING_TASK_ACTIVE_ITEMS_INCOMPLETE: statuses.PAUSING,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_PAUSED: statuses.PAUSED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.PAUSED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_COMPLETED: statuses.PAUSED,
        events.ACTION_PAUSING: statuses.PAUSING,
        events.ACTION_PAUSED: statuses.PAUSED,
        events.ACTION_PAUSED_TASK_ACTIVE_ITEMS_INCOMPLETE: statuses.PAUSING,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_PAUSED: statuses.PAUSED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.PAUSED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_COMPLETED: statuses.PAUSED,
        events.ACTION_CANCELING: statuses.CANCELING,
        events.ACTION_CANCELED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_ACTIVE_ITEMS_INCOMPLETE: statuses.CANCELING,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_COMPLETED: statuses.CANCELED,
        events.ACTION_FAILED: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_PAUSED: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_COMPLETED: statuses.FAILED,
        events.ACTION_EXPIRED: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_PAUSED: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_COMPLETED: statuses.FAILED,
        events.ACTION_ABANDONED: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_PAUSED: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_COMPLETED: statuses.FAILED,
        events.ACTION_SUCCEEDED: statuses.SUCCEEDED,
        events.ACTION_SUCCEEDED_TASK_ACTIVE_ITEMS_INCOMPLETE: statuses.RUNNING,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_PAUSED: statuses.PAUSED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.RUNNING,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_COMPLETED: statuses.SUCCEEDED,
        events.WORKFLOW_PAUSING_TASK_ACTIVE_ITEMS_INCOMPLETE: statuses.PAUSING,
        events.WORKFLOW_PAUSING_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.PAUSED,
        events.WORKFLOW_PAUSED_TASK_ACTIVE_ITEMS_INCOMPLETE: statuses.PAUSING,
        events.WORKFLOW_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.PAUSED,
        events.WORKFLOW_CANCELING_TASK_ACTIVE_ITEMS_INCOMPLETE: statuses.CANCELING,
        events.WORKFLOW_CANCELING_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.WORKFLOW_CANCELED_TASK_ACTIVE_ITEMS_INCOMPLETE: statuses.CANCELING,
        events.WORKFLOW_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
    },
    statuses.PENDING: {
        events.ACTION_CANCELED: statuses.CANCELED,
        events.ACTION_FAILED: statuses.FAILED,
        events.ACTION_EXPIRED: statuses.FAILED,
        events.ACTION_ABANDONED: statuses.FAILED,
        events.ACTION_SUCCEEDED: statuses.SUCCEEDED,
    },
    statuses.PAUSING: {
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_PAUSED: statuses.PAUSED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.PAUSED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_COMPLETED: statuses.PAUSED,
        events.ACTION_PAUSED: statuses.PAUSED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_PAUSED: statuses.PAUSED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.PAUSED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_COMPLETED: statuses.PAUSED,
        events.ACTION_RESUMING: statuses.RESUMING,
        events.ACTION_CANCELING: statuses.CANCELING,
        events.ACTION_CANCELED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_COMPLETED: statuses.CANCELED,
        events.ACTION_FAILED: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_PAUSED: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_COMPLETED: statuses.FAILED,
        events.ACTION_EXPIRED: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_PAUSED: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_COMPLETED: statuses.FAILED,
        events.ACTION_ABANDONED: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_PAUSED: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_COMPLETED: statuses.FAILED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_PAUSED: statuses.PAUSED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_FAILED: statuses.FAILED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.PAUSED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_COMPLETED: statuses.SUCCEEDED,
    },
    statuses.PAUSED: {
        events.ACTION_RUNNING: statuses.RUNNING,
        events.ACTION_RESUMING: statuses.RESUMING,
        events.ACTION_CANCELING: statuses.CANCELING,
        events.ACTION_CANCELED: statuses.CANCELED,
    },
    statuses.RESUMING: {
        events.ACTION_RUNNING: statuses.RUNNING,
        events.ACTION_CANCELING: statuses.CANCELING,
        events.ACTION_CANCELED: statuses.CANCELED,
    },
    statuses.RETRYING: {
        events.ACTION_RUNNING: statuses.RUNNING,
        events.ACTION_CANCELING: statuses.CANCELING,
        events.ACTION_CANCELED: statuses.CANCELED,
        events.WORKFLOW_CANCELING: statuses.CANCELED,
        events.WORKFLOW_CANCELED: statuses.CANCELED,
    },
    statuses.CANCELING: {
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_PENDING_TASK_DORMANT_ITEMS_COMPLETED: statuses.CANCELED,
        events.ACTION_PAUSED: statuses.CANCELED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_PAUSED_TASK_DORMANT_ITEMS_COMPLETED: statuses.CANCELED,
        events.ACTION_CANCELED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_CANCELED_TASK_DORMANT_ITEMS_COMPLETED: statuses.CANCELED,
        events.ACTION_FAILED: statuses.FAILED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_FAILED_TASK_DORMANT_ITEMS_COMPLETED: statuses.CANCELED,
        events.ACTION_EXPIRED: statuses.FAILED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_EXPIRED_TASK_DORMANT_ITEMS_COMPLETED: statuses.CANCELED,
        events.ACTION_ABANDONED: statuses.FAILED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_ABANDONED_TASK_DORMANT_ITEMS_COMPLETED: statuses.CANCELED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_PAUSED: statuses.CANCELED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_CANCELED: statuses.CANCELED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_FAILED: statuses.CANCELED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_INCOMPLETE: statuses.CANCELED,
        events.ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_COMPLETED: statuses.SUCCEEDED,
    },
    statuses.CANCELED: {},
    statuses.SUCCEEDED: {events.TASK_RETRY_REQUESTED: statuses.RETRYING},
    statuses.FAILED: {events.TASK_RETRY_REQUESTED: statuses.RETRYING},
}


class TaskStateMachine(object):
    @classmethod
    def is_transition_valid(cls, old_status, new_status):
        if old_status is None:
            old_status = "null"

        if new_status is None:
            new_status = "null"

        if not statuses.is_valid(old_status):
            raise exc.InvalidStatus(old_status)

        if not statuses.is_valid(new_status):
            raise exc.InvalidStatus(new_status)

        if old_status not in TASK_STATE_MACHINE_DATA.keys():
            return False

        if old_status == new_status or new_status in TASK_STATE_MACHINE_DATA[old_status].values():
            return True

        return False

    @classmethod
    def add_context_to_action_event(cls, workflow_state, task_id, task_route, ac_ex_event):
        return ac_ex_event.name

    @classmethod
    def process_action_event(cls, workflow_state, task_state, ac_ex_event):
        # Check if event is valid.
        if ac_ex_event.name not in events.ACTION_EXECUTION_EVENTS + events.ENGINE_OPERATION_EVENTS:
            raise exc.InvalidEvent(ac_ex_event.name)

        # Append additional task context to the event.
        event_name = cls.add_context_to_action_event(
            workflow_state, task_state["id"], task_state["route"], ac_ex_event
        )

        # Identify current task status.
        current_task_status = task_state.get("status", statuses.UNSET)

        if current_task_status is None:
            current_task_status = statuses.UNSET

        if current_task_status not in statuses.ALL_STATUSES:
            raise exc.InvalidStatus(current_task_status)

        if current_task_status not in TASK_STATE_MACHINE_DATA:
            raise exc.InvalidTaskStatusTransition(current_task_status, event_name)

        # If no transition is identified, then there is no status change.
        if event_name not in TASK_STATE_MACHINE_DATA[current_task_status]:
            return

        new_task_status = TASK_STATE_MACHINE_DATA[current_task_status][event_name]

        # Assign new status to the task flow entry.
        task_state["status"] = new_task_status

    @classmethod
    def add_context_to_task_item_event(cls, workflow_state, task_id, task_route, ac_ex_event):
        action_event = ac_ex_event.name

        requirements = [
            statuses.RESUMING,
            statuses.PENDING,
            statuses.PAUSED,
            statuses.SUCCEEDED,
            statuses.FAILED,
            statuses.EXPIRED,
            statuses.ABANDONED,
            statuses.CANCELED,
        ]

        if ac_ex_event.status in requirements:
            # Make a copy of the items and remove current item under evaluation.
            staged_task = workflow_state.get_staged_task(task_id, task_route)
            items = json_util.deepcopy(staged_task["items"])
            del items[ac_ex_event.item_id]
            items_status = [item.get("status", statuses.UNSET) for item in items]

            # Assess various situations.
            active = list(filter(lambda x: x in statuses.ACTIVE_STATUSES, items_status))
            incomplete = list(filter(lambda x: x not in statuses.COMPLETED_STATUSES, items_status))
            paused = list(filter(lambda x: x in [statuses.PENDING, statuses.PAUSED], items_status))
            canceled = list(filter(lambda x: x == statuses.CANCELED, items_status))
            failed = list(filter(lambda x: x in statuses.ABENDED_STATUSES, items_status))

            # Attach info on whether task is still active or dormant.
            action_event += "_task_active" if active else "_task_dormant"

            # Attach info on whether there are paused execution on the items and return.
            if not active and paused:
                return action_event + "_items_paused"

            # Attach info on whether there are canceled execution on the items and return.
            if not active and canceled:
                return action_event + "_items_canceled"

            # Attach info on whether there are failed execution on the items and return.
            if not active and failed:
                return action_event + "_items_failed"

            # If items are not paused, canceled, or failed, then attach info on whether
            # there are items that are not in one of the completed statuses.
            return action_event + ("_items_incomplete" if incomplete else "_items_completed")

        return action_event

    @classmethod
    def process_task_item_event(cls, workflow_state, task_state, ac_ex_event):
        # Check if event is valid.
        if ac_ex_event.name not in events.ACTION_EXECUTION_EVENTS + events.ENGINE_OPERATION_EVENTS:
            raise exc.InvalidEvent(ac_ex_event.name)

        # Append additional task context to the event.
        event_name = cls.add_context_to_task_item_event(
            workflow_state, task_state["id"], task_state["route"], ac_ex_event
        )

        # Identify current task status.
        current_task_status = task_state.get("status", statuses.UNSET)

        if current_task_status is None:
            current_task_status = statuses.UNSET

        if current_task_status not in statuses.ALL_STATUSES:
            raise exc.InvalidStatus(current_task_status)

        if current_task_status not in TASK_STATE_MACHINE_DATA:
            raise exc.InvalidTaskStatusTransition(current_task_status, event_name)

        # If no transition is identified, then there is no status change.
        if event_name not in TASK_STATE_MACHINE_DATA[current_task_status]:
            return

        new_task_status = TASK_STATE_MACHINE_DATA[current_task_status][event_name]

        # Assign new status to the task flow entry.
        task_state["status"] = new_task_status

    @classmethod
    def add_context_to_workflow_event(cls, workflow_state, task_id, task_route, wf_ex_event):
        workflow_event = wf_ex_event.name
        requirements = statuses.PAUSE_STATUSES + statuses.CANCEL_STATUSES
        staged_task = workflow_state.get_staged_task(task_id, task_route)

        if wf_ex_event.status in requirements and staged_task and "items" in staged_task:
            items_status = [item.get("status", statuses.UNSET) for item in staged_task["items"]]
            active = list(filter(lambda x: x in statuses.ACTIVE_STATUSES, items_status))
            incomplete = list(filter(lambda x: x not in statuses.COMPLETED_STATUSES, items_status))
            workflow_event += "_task_active" if active else "_task_dormant"
            workflow_event += "_items_incomplete" if incomplete else "_items_completed"

        return workflow_event

    @classmethod
    def process_workflow_event(cls, workflow_state, task_state, wf_ex_event):
        # Check if event is valid.
        if wf_ex_event.name not in events.WORKFLOW_EXECUTION_EVENTS:
            raise exc.InvalidEvent(wf_ex_event.name)

        # Append additional task context to the event.
        event_name = cls.add_context_to_workflow_event(
            workflow_state, task_state["id"], task_state["route"], wf_ex_event
        )

        # Identify current task status.
        current_task_status = task_state.get("status", statuses.UNSET)

        if current_task_status is None:
            current_task_status = statuses.UNSET

        if current_task_status not in statuses.ALL_STATUSES:
            raise exc.InvalidStatus(current_task_status)

        if current_task_status not in TASK_STATE_MACHINE_DATA:
            raise exc.InvalidTaskStatusTransition(current_task_status, event_name)

        # If no transition is identified, then there is no status change.
        if event_name not in TASK_STATE_MACHINE_DATA[current_task_status]:
            return

        new_task_status = TASK_STATE_MACHINE_DATA[current_task_status][event_name]

        # Assign new status to the task flow entry.
        task_state["status"] = new_task_status

    @classmethod
    def process_event(cls, workflow_state, task_state, event):
        if isinstance(event, events.WorkflowExecutionEvent):
            cls.process_workflow_event(workflow_state, task_state, event)
            return

        if isinstance(event, events.TaskItemActionExecutionEvent):
            cls.process_task_item_event(workflow_state, task_state, event)
            return

        if isinstance(event, events.ActionExecutionEvent):
            cls.process_action_event(workflow_state, task_state, event)
            return

        if isinstance(event, events.EngineOperationEvent):
            cls.process_action_event(workflow_state, task_state, event)
            return

        raise exc.InvalidEventType(type(event), event.name)


class WorkflowStateMachine(object):
    @classmethod
    def is_transition_valid(cls, old_status, new_status):
        if old_status is None:
            old_status = "null"

        if new_status is None:
            new_status = "null"

        if not statuses.is_valid(old_status):
            raise exc.InvalidStatus(old_status)

        if not statuses.is_valid(new_status):
            raise exc.InvalidStatus(new_status)

        if old_status not in WORKFLOW_STATE_MACHINE_DATA.keys():
            return False

        if (
            old_status == new_status
            or new_status in WORKFLOW_STATE_MACHINE_DATA[old_status].values()
        ):
            return True

        return False

    @classmethod
    def add_context_to_task_event(cls, workflow_state, tk_ex_event):
        # Identify current workflow status.
        task_event = tk_ex_event.name
        task_id = getattr(tk_ex_event, "task_id", None)
        task_route = getattr(tk_ex_event, "route", None)
        has_next_tasks = workflow_state.has_next_tasks(task_id, task_route)
        has_active_tasks = workflow_state.has_active_tasks

        # Mark task remediated if task is in abended statuses and there are transitions.
        if tk_ex_event.status in statuses.ABENDED_STATUSES and has_next_tasks:
            task_event = events.TASK_REMEDIATED

        # For certain events like cancel and pause, whether there are tasks in active
        # status determine whether the workflow reached final status or still in progress.
        # For example, if the workflow is being canceled and there are other active
        # tasks, the workflow should be set to canceling.
        if task_event in events.TASK_CONDITIONAL_EVENTS:
            task_event += "_workflow_active" if has_active_tasks else "_workflow_dormant"

        # When a task succeeded, additional information need to be included in the
        # event to determine whether the workflow is still running, completed, or
        # it needs to be paused or canceled. Cancelation has precedence over other
        # status, follow by pause, running, then completion.

        if not task_event.startswith(events.TASK_SUCCEEDED) and not task_event.startswith(
            events.TASK_REMEDIATED
        ):
            return task_event

        if workflow_state.has_canceling_tasks or workflow_state.has_canceled_tasks:
            return task_event + "_canceled"

        if workflow_state.has_pausing_tasks or workflow_state.has_paused_tasks:
            return task_event + "_paused"

        if workflow_state.has_staged_tasks or has_next_tasks:
            return task_event + "_incomplete"

        return task_event + "_completed"

    @classmethod
    def process_task_event(cls, workflow_state, tk_ex_event):
        # Append additional workflow context to the event.
        event_name = cls.add_context_to_task_event(workflow_state, tk_ex_event)

        # Check if event is valid.
        if event_name not in events.TASK_EXECUTION_EVENTS:
            raise exc.InvalidEvent(event_name)

        # Capture current workflow status.
        current_workflow_status = workflow_state.status
        new_workflow_status = current_workflow_status

        # Check if the current workflow status can be transitioned.
        if current_workflow_status not in WORKFLOW_STATE_MACHINE_DATA:
            raise exc.InvalidWorkflowStatusTransition(current_workflow_status, event_name)

        # If the current workflow status can be transitioned and there is no match on the
        # event, then there is not status transition.
        if event_name not in WORKFLOW_STATE_MACHINE_DATA[current_workflow_status]:
            return

        new_workflow_status = WORKFLOW_STATE_MACHINE_DATA[current_workflow_status][event_name]

        # Assign new workflow status if there is change.
        if current_workflow_status != new_workflow_status:
            workflow_state.status = new_workflow_status

        # If the final workflow status here is completed, then ensure there is no unreachable
        # barrier task(s). A barrier task is unreachable if the workflow is completed but then one
        # or more criteria for the task is satisified. In this case, log the task and fail the
        # workflow to notify that the execution is incomplete but unable to proceed.
        if workflow_state.status in statuses.COMPLETED_STATUSES:
            unreachable_barriers = workflow_state.get_unreachable_barriers()

            # If there are unreachable barrier tasks, then change workflow status to failed
            # and write an error log for each case.
            if unreachable_barriers:
                workflow_state.status = statuses.FAILED

                for entry in unreachable_barriers:
                    e = exc.UnreachableJoinError(entry["id"], entry["route"])
                    workflow_state.conductor.log_error(e, task_id=entry["id"], route=entry["route"])

    @classmethod
    def add_context_to_workflow_event(cls, workflow_state, wf_ex_event):
        # Identify current workflow status.
        workflow_event = wf_ex_event.name
        has_active_tasks = workflow_state.has_active_tasks

        # For certain events like cancel and pause, whether there are tasks in active
        # status determine whether the workflow reached final status or still in progress.
        # For example, if the workflow is being canceled and there are other active
        # tasks, the workflow should be set to canceling.
        if wf_ex_event.status in statuses.PAUSE_STATUSES + statuses.CANCEL_STATUSES:
            workflow_event += "_workflow_active" if has_active_tasks else "_workflow_dormant"

        # If the workflow is paused and on resume, check whether it is already completed.
        if (
            workflow_state.status == statuses.PAUSED
            and wf_ex_event.status in [statuses.RUNNING, statuses.RESUMING]
            and not workflow_state.has_active_tasks
            and not workflow_state.has_staged_tasks
            and not workflow_state.has_paused_tasks
        ):
            workflow_event += "_workflow_completed"

        return workflow_event

    @classmethod
    def process_workflow_event(cls, workflow_state, wf_ex_event):
        # Append additional workflow context to the event.
        event_name = cls.add_context_to_workflow_event(workflow_state, wf_ex_event)

        # Check if event is valid.
        if event_name not in events.WORKFLOW_EXECUTION_EVENTS:
            raise exc.InvalidEvent(event_name)

        # Capture current workflow status.
        current_workflow_status = workflow_state.status
        new_workflow_status = current_workflow_status

        # Check if the current workflow status can be transitioned.
        if current_workflow_status not in WORKFLOW_STATE_MACHINE_DATA:
            raise exc.InvalidWorkflowStatusTransition(current_workflow_status, event_name)

        # If the current workflow status can be transitioned and there is no match on the
        # event, then there is not status transition.
        if event_name not in WORKFLOW_STATE_MACHINE_DATA[current_workflow_status]:
            return

        new_workflow_status = WORKFLOW_STATE_MACHINE_DATA[current_workflow_status][event_name]

        # Assign new workflow status if there is change.
        if current_workflow_status != new_workflow_status:
            workflow_state.status = new_workflow_status

    @classmethod
    def process_event(cls, workflow_state, event):
        if isinstance(event, events.WorkflowExecutionEvent):
            cls.process_workflow_event(workflow_state, event)
            return

        if isinstance(event, events.TaskExecutionEvent):
            cls.process_task_event(workflow_state, event)
            return

        raise exc.InvalidEventType(type(event), event.name)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import six
from six.moves import queue
import yaml

from orquesta import events
from orquesta import exceptions as exc
from orquesta.expressions import base as expr_base
from orquesta.specs.native.v1 import base as native_v1_specs
from orquesta.specs import types as spec_types
from orquesta.utils import context as ctx_util
from orquesta.utils import dictionary as dict_util
from orquesta.utils import jsonify as json_util
from orquesta.utils import parameters as args_util


LOG = logging.getLogger(__name__)

RESERVED_TASK_NAMES = list(events.ENGINE_EVENT_MAP.keys())


def instantiate(definition):
    return WorkflowSpec(definition)


def deserialize(data):
    return WorkflowSpec.deserialize(data)


class TaskTransitionSpec(native_v1_specs.Spec):
    _schema = {
        "type": "object",
        "properties": {
            "when": spec_types.NONEMPTY_STRING,
            "publish": {"oneOf": [spec_types.NONEMPTY_STRING, spec_types.UNIQUE_ONE_KEY_DICT_LIST]},
            "do": {"oneOf": [spec_types.NONEMPTY_STRING, spec_types.UNIQUE_STRING_LIST]},
        },
        "additionalProperties": False,
    }

    _context_evaluation_sequence = ["when", "publish", "do"]

    _context_inputs = ["publish"]

    def __init__(self, *args, **kwargs):
        super(TaskTransitionSpec, self).__init__(*args, **kwargs)

        publish_spec = getattr(self, "publish", None)

        if publish_spec and isinstance(publish_spec, six.string_types):
            self.publish = args_util.parse_inline_params(publish_spec or str())

        do_spec = getattr(self, "do", None)

        if not do_spec:
            self.do = "continue"


class TaskTransitionSequenceSpec(native_v1_specs.SequenceSpec):
    _schema = {"type": "array", "items": TaskTransitionSpec}


class ItemizedSpec(native_v1_specs.Spec):
    _items_regex = (
        # Regular expression in the form "x, y, z, ... in <expression>"
        # or "x in <expression>" with optional space(s) on both end.
        r"^(\s+)?({expr})(\s+)?$|^(\s+)?((\w+,\s?|\s+)+)?(\w+)\s+in\s+({expr})(\s+)?$".format(
            expr="|".join(expr_base.get_statement_regexes().values())
        )
    )

    _schema = {
        "type": "object",
        "properties": {
            "items": {"type": "string", "minLength": 1, "pattern": _items_regex},
            "concurrency": spec_types.STRING_OR_POSITIVE_INTEGER,
        },
        "additionalProperties": False,
    }

    _context_evaluation_sequence = ["items", "concurrency"]


class TaskRetrySpec(native_v1_specs.Spec):
    _schema = {
        "type": "object",
        "properties": {
            "when": spec_types.NONEMPTY_STRING,
            # The number of times to retry.
            # 0 = do not retry
            # n = max number of times to retry
            "count": spec_types.STRING_OR_POSITIVE_INTEGER,
            "delay": spec_types.STRING_OR_POSITIVE_INTEGER,
        },
        "required": ["count"],
        "additionalProperties": False,
    }

    _context_evaluation_sequence = ["when", "count", "delay"]


class TaskSpec(native_v1_specs.Spec):
    _schema = {
        "type": "object",
        "properties": {
            "delay": spec_types.STRING_OR_POSITIVE_INTEGER,
            "join": {"oneOf": [{"enum": ["all"]}, spec_types.POSITIVE_INTEGER]},
            "with": ItemizedSpec,
            "action": spec_types.NONEMPTY_STRING,
            "input": {"oneOf": [spec_types.NONEMPTY_STRING, spec_types.NONEMPTY_DICT]},
            "retry": TaskRetrySpec,
            "next": TaskTransitionSequenceSpec,
        },
        "additionalProperties": False,
    }

    _context_evaluation_sequence = ["delay", "with", "action", "input", "next"]

    def __init__(self, *args, **kwargs):
        super(TaskSpec, self).__init__(*args, **kwargs)

        action_spec = getattr(self, "action", str())
        input_spec = args_util.parse_inline_params(action_spec, preserve_order=False)

        if input_spec:
            self.action = action_spec[: action_spec.index(" ")]
            self.input = input_spec

    def has_items(self):
        return hasattr(self, "with") and getattr(self, "with", None) is not None

    def get_items_spec(self):
        return getattr(self, "with", None)

    def has_join(self):
        return hasattr(self, "join") and self.join

    def has_retry(self):
        return hasattr(self, "retry") and self.retry

    def render(self, in_ctx):
        action_specs = []

        if not self.has_items():
            action_spec = {
                "action": expr_base.evaluate(self.action, in_ctx),
                "input": expr_base.evaluate(getattr(self, "input", {}), in_ctx),
            }

            action_specs.append(action_spec)
        else:
            items_spec = self.get_items_spec()

            if " in " not in items_spec.items:
                items_expr = items_spec.items.strip()
            else:
                start_idx = items_spec.items.index(" in ") + 4
                items_expr = items_spec.items[start_idx:].strip()

            items = expr_base.evaluate(items_expr, in_ctx)

            if not isinstance(items, list):
                raise TypeError('The value of "%s" is not type of list.' % items_expr)

            item_keys = (
                None
                if " in " not in items_spec.items
                else items_spec.items[: items_spec.items.index(" in ")].replace(" ", "").split(",")
            )

            for idx, item in enumerate(items):
                if item_keys and (isinstance(item, tuple) or isinstance(item, list)):
                    item = dict(zip(item_keys, list(item)))
                elif item_keys and len(item_keys) == 1:
                    item = {item_keys[0]: item}

                item_ctx_value = ctx_util.set_current_item(in_ctx, item)

                action_spec = {
                    "action": expr_base.evaluate(self.action, item_ctx_value),
                    "input": expr_base.evaluate(getattr(self, "input", {}), item_ctx_value),
                    "item_id": idx,
                }

                action_specs.append(action_spec)

        return self, action_specs

    def finalize_context(self, next_task_name, task_transition_meta, in_ctx):
        rolling_ctx = json_util.deepcopy(in_ctx)
        new_ctx = {}
        errors = []

        task_transition_specs = getattr(self, "next") or []
        task_transition_spec = task_transition_specs[task_transition_meta[3]["ref"]]
        next_task_names = getattr(task_transition_spec, "do") or []

        if next_task_name in next_task_names:
            for task_publish_spec in getattr(task_transition_spec, "publish") or {}:
                var_name = list(task_publish_spec.items())[0][0]
                default_var_value = list(task_publish_spec.items())[0][1]

                try:
                    rendered_var_value = expr_base.evaluate(default_var_value, rolling_ctx)
                    rolling_ctx[var_name] = rendered_var_value
                    new_ctx[var_name] = rendered_var_value
                except exc.ExpressionEvaluationException as e:
                    errors.append(e)

        out_ctx = dict_util.merge_dicts(in_ctx, new_ctx, overwrite=True)

        for key in list(out_ctx.keys()):
            if key.startswith("__"):
                out_ctx.pop(key)

        return out_ctx, new_ctx, errors


class TaskMappingSpec(native_v1_specs.MappingSpec):
    _schema = {"type": "object", "minProperties": 1, "patternProperties": {r"^\w+$": TaskSpec}}

    def has_tasks(self):
        return len(self.keys()) > 0

    def has_task(self, task_name):
        if task_name in RESERVED_TASK_NAMES:
            return True

        return task_name in self

    def get_task(self, task_name):
        if task_name in RESERVED_TASK_NAMES:
            return TaskSpec({"name": task_name})

        return self[task_name]

    def get_next_tasks(self, task_name, *args, **kwargs):
        task_spec = self.get_task(task_name)

        next_tasks = []

        task_transitions = getattr(task_spec, "next") or []

        for task_transition_item_idx, task_transition in enumerate(task_transitions):
            condition = getattr(task_transition, "when") or None
            next_task_names = getattr(task_transition, "do") or []

            if isinstance(next_task_names, six.string_types):
                next_task_names = [x.strip() for x in next_task_names.split(",")]

            for next_task_name in next_task_names:
                next_tasks.append((next_task_name, condition, task_transition_item_idx))

        return sorted(next_tasks, key=lambda x: x[0])

    def get_prev_tasks(self, task_name, *args, **kwargs):
        prev_tasks = []

        for name, task_spec in six.iteritems(self):
            for next_task in self.get_next_tasks(name):
                if task_name == next_task[0]:
                    prev_tasks.append((name, next_task[1], next_task[2]))

        return sorted(prev_tasks, key=lambda x: x[0])

    def get_start_tasks(self):
        start_tasks = [
            (task_name, None, None)
            for task_name in self.keys()
            if not self.get_prev_tasks(task_name)
        ]

        return sorted(start_tasks, key=lambda x: x[0])

    def is_join_task(self, task_name):
        task_spec = self.get_task(task_name)

        return getattr(task_spec, "join", None) is not None

    def is_split_task(self, task_name):
        return not self.is_join_task(task_name) and len(self.get_prev_tasks(task_name)) > 1

    def in_cycle(self, task_name):
        traversed = []
        q = queue.Queue()

        for task in self.get_next_tasks(task_name):
            q.put(task[0])

        while not q.empty():
            next_task_name = q.get()

            # If the next task matches the original task, then it's in a loop.
            if next_task_name == task_name:
                return True

            # If the next task has already been traversed but didn't match the
            # original task, then there's a loop but the original task is not
            # in the loop.
            if next_task_name in traversed:
                continue

            for task in self.get_next_tasks(next_task_name):
                q.put(task[0])

            traversed.append(next_task_name)

        return False

    def has_cycles(self):
        for task_name, task_spec in six.iteritems(self):
            if self.in_cycle(task_name):
                return True

        return False

    def detect_actionless_with_items(self, parent=None):
        result = []

        # Identify with items task with no action defined.
        for task_name, task_spec in six.iteritems(self):
            if task_spec.has_items() and not task_spec.action:
                message = "The action property is required for with items task."
                spec_path = parent.get("spec_path") + "." + task_name
                schema_path = parent.get("schema_path") + ".patternProperties.^\\w+$"
                entry = {"message": message, "spec_path": spec_path, "schema_path": schema_path}
                result.append(entry)

        return result

    def detect_reserved_names(self, parent=None):
        result = []

        # Identify use of reserved words in task names.
        for task_name, task_spec in six.iteritems(self):
            if task_name in RESERVED_TASK_NAMES:
                message = 'The task name "%s" is reserved with special function.' % task_name
                spec_path = parent.get("spec_path") + "." + task_name
                schema_path = parent.get("schema_path") + ".patternProperties.^\\w+$"
                entry = {"message": message, "spec_path": spec_path, "schema_path": schema_path}
                result.append(entry)

        return result

    def detect_start_tasks(self, parent=None):
        result = []

        if self.has_tasks() and not self.get_start_tasks():
            entry = {
                "message": (
                    "Unable to identify any tasks to start the workflow. If the workflow has a "
                    "loop, please add a task that serves as an entry point (or lead) into the loop."
                ),
                "spec_path": parent.get("spec_path"),
                "schema_path": parent.get("schema_path"),
            }

            result.append(entry)

        return result

    def detect_undefined_tasks(self, parent=None):
        # Identify the undefined task in task transitions.
        result = []
        traversed = []
        q = queue.Queue()

        for task in self.get_start_tasks():
            q.put(task[0])

        while not q.empty():
            task_name = q.get()
            traversed.append(task_name)

            # Identify the next set of tasks and related transition specs.
            # The get_next_tasks function is not used here because it doesn't
            # provide the specific info required to identify spec/schema paths.
            task_spec = self.get_task(task_name)
            task_transition_specs = getattr(task_spec, "next") or []
            spec_path = parent.get("spec_path") + "." + task_name
            schema_path = parent.get("schema_path") + ".patternProperties.^\\w+$"

            for i in range(0, len(task_transition_specs)):
                task_transition_spec = task_transition_specs[i]
                next_task_names = getattr(task_transition_spec, "do") or []

                if isinstance(next_task_names, six.string_types):
                    next_task_names = [x.strip() for x in next_task_names.split(",")]

                for next_task_name in next_task_names:
                    # If the next task has already been traversed, then skip.
                    if next_task_name in traversed:
                        continue

                    if self.has_task(next_task_name):
                        if next_task_name not in RESERVED_TASK_NAMES + traversed:
                            q.put(next_task_name)
                    else:
                        entry = {
                            "message": 'The task "%s" is not defined.' % next_task_name,
                            "spec_path": spec_path + ".next[" + str(i) + "].do",
                            "schema_path": schema_path + ".properties.next.items.properties.do",
                        }

                        result.append(entry)

        return result

    def detect_unreachable_tasks(self, parent=None):
        # Identify tasks that are not reachable.
        result = []
        staging = {}
        q = queue.Queue()

        # Traverse the tasks spec and prep data for evaluation.
        for task_name, condition, task_transition_item_idx in self.get_start_tasks():
            q.put((None, task_name, []))

        while not q.empty():
            prev_task_name, task_name, splits = q.get()

            # If the task is undefined, then move on.
            if task_name not in self:
                continue

            # Determine if the task is a split task and if it is in a cycle. If the task is a
            # split task, keep track of where the split(s) occurs.
            if self.is_split_task(task_name) and not self.in_cycle(task_name):
                splits.append(task_name)

            if task_name not in staging:
                staging[task_name] = {
                    "spec_path": parent.get("spec_path") + "." + task_name,
                    "schema_path": parent.get("schema_path") + ".patternProperties.^\\w+$",
                    "join": self.is_join_task(task_name),
                    "splits": [],
                    "prev": [],
                }

            staging[task_name]["splits"] = list(set(staging[task_name]["splits"]) | set(splits))

            if prev_task_name:
                inbounds = list(set(staging[task_name]["prev"]) | set([prev_task_name]))
                staging[task_name]["prev"] = inbounds

            next_tasks = self.get_next_tasks(task_name)

            for next_task_name, condition, task_transition_item_idx in next_tasks:
                if next_task_name not in staging or not self.in_cycle(next_task_name):
                    q.put((task_name, next_task_name, list(splits)))

        # Use the prepped data to identify tasks that are not reachable.
        for task_name, meta in six.iteritems(staging):
            if not meta["join"]:
                meta["reachable"] = True
                continue

            for prev_task_name in meta["prev"]:
                meta["reachable"] = meta["splits"] == staging[prev_task_name]["splits"]

                if not meta["reachable"]:
                    msg = (
                        'The join task "%s" is unreachable. A join task is determined to be '
                        "unreachable if there are nested forks from multi-referenced tasks "
                        "that join on the said task. This is ambiguous to the workflow engine "
                        "because it does not know at which level should the join occurs."
                    )

                    entry = {
                        "message": msg % task_name,
                        "spec_path": meta["spec_path"],
                        "schema_path": meta["schema_path"],
                    }

                    result.append(entry)

                    break

        return result

    def inspect_semantics(self, parent=None):
        result = self.detect_reserved_names(parent=parent)
        result.extend(self.detect_start_tasks(parent=parent))
        result.extend(self.detect_undefined_tasks(parent=parent))
        result.extend(self.detect_unreachable_tasks(parent=parent))
        result.extend(self.detect_actionless_with_items(parent=parent))

        return result

    def inspect_context(self, parent=None):
        ctxs = {}
        errors = []
        traversed = []
        parent_ctx = parent.get("ctx", []) if parent else []
        rolling_ctx = list(set(parent_ctx))
        q = queue.Queue()

        for task in self.get_start_tasks():
            q.put((task[0], json_util.deepcopy(rolling_ctx)))

        while not q.empty():
            task_name, task_ctx = q.get()
            traversed.append(task_name)

            if not task_ctx:
                task_ctx = ctxs.get(task_name, [])

            task_spec = self.get_task(task_name)

            spec_path = parent.get("spec_path") + "." + task_name
            schema_path = parent.get("schema_path") + ".patternProperties.^\\w+$"

            task_parent = {"ctx": task_ctx, "spec_path": spec_path, "schema_path": schema_path}

            result = task_spec.inspect_context(parent=task_parent)
            errors.extend(result[0])
            task_ctx = list(set(task_ctx + result[1]))
            rolling_ctx = list(set(rolling_ctx + task_ctx))

            # Identify the next set of tasks and related transition specs.
            transitions = []
            task_transition_specs = getattr(task_spec, "next") or []

            for i in range(0, len(task_transition_specs)):
                task_transition_spec = task_transition_specs[i]
                next_task_names = getattr(task_transition_spec, "do") or []

                if not next_task_names:
                    transitions.append((None, task_transition_spec, str(i)))
                    continue

                if isinstance(next_task_names, six.string_types):
                    next_task_names = [x.strip() for x in next_task_names.split(",")]

                for next_task_name in next_task_names:
                    entry = (next_task_name, task_transition_spec, str(i))
                    transitions.append(entry)

            for entry in transitions:
                next_task_name = entry[0]
                task_transition_spec = entry[1]
                seq_num = entry[2]

                parent_ctx = {
                    "ctx": task_ctx,
                    "spec_path": spec_path + ".next[" + seq_num + "]",
                    "schema_path": schema_path + ".properties.next.items",
                }

                result = task_transition_spec.inspect_context(parent_ctx)
                errors.extend(result[0])
                branch_ctx = list(set(task_ctx + result[1]))

                if (
                    not next_task_name
                    or next_task_name in traversed
                    or not self.has_task(next_task_name)
                ):
                    continue

                next_task_spec = self.get_task(next_task_name)

                if not next_task_spec.has_join():
                    q.put((next_task_name, branch_ctx))
                else:
                    next_task_ctx = ctxs.get(next_task_name, [])
                    ctxs[next_task_name] = list(set(next_task_ctx + branch_ctx))
                    q.put((next_task_name, None))

        return (errors, rolling_ctx)


class WorkflowSpec(native_v1_specs.Spec):
    _schema = {
        "type": "object",
        "properties": {
            "vars": spec_types.UNIQUE_ONE_KEY_DICT_LIST,
            "input": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
            "output": spec_types.UNIQUE_ONE_KEY_DICT_LIST,
            "tasks": TaskMappingSpec,
        },
        "required": ["tasks"],
        "additionalProperties": False,
    }

    _context_evaluation_sequence = ["input", "vars", "tasks", "output"]

    _context_inputs = ["input", "vars", "output"]

    def __init__(self, spec, name=None, member=False):
        if not spec:
            raise ValueError("The spec cannot be type of None.")

        spec = (
            yaml.safe_load(spec)
            if not isinstance(spec, dict) and not isinstance(spec, list)
            else spec
        )

        # Resolve shorthand inline with items.
        if "tasks" in spec and isinstance(spec["tasks"], dict) and spec["tasks"]:
            for task_name, task_spec in six.iteritems(spec["tasks"]):
                if "with" in task_spec and isinstance(task_spec["with"], six.string_types):
                    task_spec["with"] = {"items": task_spec["with"]}

        super(WorkflowSpec, self).__init__(spec, name=name, member=member)

    def render_input(self, runtime_inputs, in_ctx=None):
        rolling_ctx = json_util.deepcopy(in_ctx) if in_ctx else {}
        errors = []

        for input_spec in getattr(self, "input") or []:
            if isinstance(input_spec, dict):
                input_name = list(input_spec.items())[0][0]
                default_input_value = list(input_spec.items())[0][1]
            else:
                input_name = input_spec
                default_input_value = None

            runtime_input_value = runtime_inputs.get(input_name, default_input_value)

            try:
                rendered_input_value = expr_base.evaluate(runtime_input_value, rolling_ctx)
                rolling_ctx[input_name] = rendered_input_value
            except exc.ExpressionEvaluationException as e:
                errors.append(e)

        return rolling_ctx, errors

    def render_vars(self, in_ctx):
        rolling_ctx = json_util.deepcopy(in_ctx)
        rendered_vars = {}
        errors = []

        for var_spec in getattr(self, "vars") or []:
            var_name = list(var_spec.items())[0][0]
            default_var_value = list(var_spec.items())[0][1]

            try:
                rendered_var_value = expr_base.evaluate(default_var_value, rolling_ctx)
                rolling_ctx[var_name] = rendered_var_value
                rendered_vars[var_name] = rendered_var_value
            except exc.ExpressionEvaluationException as e:
                errors.append(e)

        return rendered_vars, errors

    def render_output(self, in_ctx):
        output_specs = getattr(self, "output") or []
        rolling_ctx = json_util.deepcopy(in_ctx)
        rendered_outputs = {}
        errors = []

        for output_spec in output_specs:
            output_name = list(output_spec.items())[0][0]
            default_output_value = list(output_spec.items())[0][1]

            try:
                rendered_output_value = expr_base.evaluate(default_output_value, rolling_ctx)
                rolling_ctx[output_name] = rendered_output_value
                rendered_outputs[output_name] = rendered_output_value
            except exc.ExpressionEvaluationException as e:
                errors.append(e)

        return rendered_outputs, errors



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.specs.native import base as test_base


class WorkflowSpecValidationTest(test_base.OrchestraWorkflowSpecTest):
    def test_success(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - when: <% succeeded() %>
                    publish: foo="bar"
                    do: task2
              task2:
                action: core.noop
                next:
                  - when: <% succeeded() %>
                    publish: bar="foo"
                    do: task3
              task3:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_empty_task_transition(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next: []
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_basic_task_transition(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - do:
                      - task2
              task2:
                action: core.noop
                next:
                  - do: task3
              task3:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_publish_in_task_transition(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - publish: foo="bar" bar="foo"
                    do: task2
              task2:
                action: std.echo
                input:
                    message: <% ctx().foo + ctx().bar %>
                next:
                  - publish:
                      - foobar: fubar
                      - fubar: foobar
                    do: task3
              task3:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_bad_when_in_task_transition(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - when:
                      - foobar
                    do: task2
              task2:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "['foobar'] is not of type 'string'",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.next.items.properties.when.type"
                    ),
                    "spec_path": "tasks.task1.next[0].when",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_publish_in_task_transition(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - publish:
                      - foobar
                    do: task2
              task2:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "['foobar'] is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.next.items.properties.publish.oneOf"
                    ),
                    "spec_path": "tasks.task1.next[0].publish",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_do_in_task_transition(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - do:
                      task2: foobar
              task2:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "{'task2': 'foobar'} is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.next.items.properties.do.oneOf"
                    ),
                    "spec_path": "tasks.task1.next[0].do",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_missing_task_list(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "'tasks' is a required property",
                    "schema_path": "required",
                    "spec_path": None,
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_empty_task_list(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks: {}
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "{} does not have enough properties",
                    "schema_path": "properties.tasks.minProperties",
                    "spec_path": "tasks",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_fail_multiple_inspection(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            vars:
              - macro: polo
            tasks:
              task1:
                action: core.local cmd=<% ctx().foobar %>
                next:
                  - when: <% succeeded() %>
                    do:
                      - task2
              task2:
                action: core.local
                input:
                    - cmd: echo <% ctx().macro %>
                next:
                  - when: <% <% succeeded() %>
                    do: task3
              task3:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "expressions": [
                {
                    "spec_path": "tasks.task2.next[0].when",
                    "expression": "<% <% succeeded() %>",
                    "message": (
                        "Parse error: unexpected '<' at position 0 of "
                        "expression '<% succeeded()'"
                    ),
                    "type": "yaql",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.next.items.properties.when"
                    ),
                }
            ],
            "context": [
                {
                    "spec_path": "tasks.task1.input",
                    "expression": "<% ctx().foobar %>",
                    "message": 'Variable "foobar" is referenced before assignment.',
                    "type": "yaql",
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                }
            ],
            "syntax": [
                {
                    "spec_path": "tasks.task2.input",
                    "message": (
                        "[{'cmd': 'echo <% ctx().macro %>'}] is "
                        "not valid under any of the given schemas"
                    ),
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$." "properties.input.oneOf"
                    ),
                }
            ],
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_missing_task_node(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - do: task2
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "semantics": [
                {
                    "message": 'The task "task2" is not defined.',
                    "spec_path": "tasks.task1.next[0].do",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.next.items.properties.do"
                    ),
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_reserved_task_name_used(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - do: noop
              noop:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "semantics": [
                {
                    "message": 'The task name "noop" is reserved with special function.',
                    "spec_path": "tasks.noop",
                    "schema_path": "properties.tasks.patternProperties.^\\w+$",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_unreachable_task(self):
        wf_def = """
            version: 1.0

            description: >
              A basic workflow that demonstrate multiple splits where a post split
              task5 is also being joined with task1 before the split. The task5 is
              not going to be unreachable because the reference at task1 is out of
              scope for the reference below task4.

            tasks:
              task1:
                action: core.noop
                next:
                  - when: <% succeeded() %>
                    do: task2, task3, task5

              # branch 1
              task2:
                action: core.noop
                next:
                  - when: <% succeeded() %>
                    do: task4

              # branch 2
              task3:
                action: core.noop
                next:
                  - when: <% succeeded() %>
                    do: task4

              # split branch
              task4:
                action: core.noop
                next:
                  - when: <% succeeded() %>
                    do: task5

              task5:
                join: all
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "semantics": [
                {
                    "message": (
                        'The join task "task5" is unreachable. A join task is determined to be '
                        "unreachable if there are nested forks from multi-referenced tasks "
                        "that join on the said task. This is ambiguous to the workflow engine "
                        "because it does not know at which level should the join occurs."
                    ),
                    "spec_path": "tasks.task5",
                    "schema_path": "properties.tasks.patternProperties.^\\w+$",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_unidentified_start_tasks(self):
        wf_def = """
            version: 1.0

            description: A sample workflow with a single task loop.

            input:
              - count: 0

            tasks:
              task1:
                action: core.noop
                next:
                  - when: <% ctx().count < 2 %>
                    publish:
                      - count: <% ctx().count + 1 %>
                    do: task1
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "semantics": [
                {
                    "message": (
                        "Unable to identify any tasks to start the workflow. "
                        "If the workflow has a loop, please add a task that "
                        "serves as an entry point (or lead) into the loop."
                    ),
                    "spec_path": "tasks",
                    "schema_path": "properties.tasks",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.composition.native import base


class SplitWorkflowComposerTest(base.OrchestraWorkflowComposerTest):
    def test_split(self):
        wf_name = "split"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4", "splits": ["task4"]},
                {"id": "task5", "splits": ["task4"]},
                {"id": "task6", "splits": ["task4"]},
                {"id": "task7", "barrier": "*", "splits": ["task4"]},
            ],
            "adjacency": [
                [
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task3", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "task5", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task6", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4", "splits": ["task4"]},
                {"id": "task5", "splits": ["task4"]},
                {"id": "task6", "splits": ["task4"]},
                {"id": "task7", "barrier": "*", "splits": ["task4"]},
            ],
            "adjacency": [
                [
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task3", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "task5", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task6", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_splits(self):
        wf_name = "splits"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4", "splits": ["task4"]},
                {"id": "task5", "splits": ["task4"]},
                {"id": "task6", "splits": ["task4"]},
                {"id": "task7", "barrier": "*", "splits": ["task4"]},
                {"id": "task8", "splits": ["task4", "task8"]},
            ],
            "adjacency": [
                [
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task3", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task8", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "task5", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task6", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task8", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4", "splits": ["task4"]},
                {"id": "task5", "splits": ["task4"]},
                {"id": "task6", "splits": ["task4"]},
                {"id": "task7", "barrier": "*", "splits": ["task4"]},
                {"id": "task8", "splits": ["task4", "task8"]},
            ],
            "adjacency": [
                [
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task3", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task8", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "task5", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task6", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task8", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_nested_splits(self):
        wf_name = "splits-nested"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4", "splits": ["task4"]},
                {"id": "task5", "splits": ["task4"]},
                {"id": "task6", "splits": ["task4"]},
                {"id": "task7", "splits": ["task4", "task7"]},
                {"id": "task8", "splits": ["task4", "task7"]},
                {"id": "task9", "splits": ["task4", "task7"]},
                {"id": "task10", "barrier": "*", "splits": ["task4", "task7"]},
            ],
            "adjacency": [
                [
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task3", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "task5", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task6", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "task8", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task9", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task10", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task10", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4", "splits": ["task4"]},
                {"id": "task5", "splits": ["task4"]},
                {"id": "task6", "splits": ["task4"]},
                {"id": "task7", "splits": ["task4", "task7"]},
                {"id": "task8", "splits": ["task4", "task7"]},
                {"id": "task9", "splits": ["task4", "task7"]},
                {"id": "task10", "barrier": "*", "splits": ["task4", "task7"]},
            ],
            "adjacency": [
                [
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task3", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task4", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "task5", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task6", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task7", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"id": "task8", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task9", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "task10", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "task10", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_very_many_splits(self):
        wf_name = "splits-very-many"

        expected_wf_graph = {
            "directed": True,
            "graph": [],
            "nodes": [
                {"id": "task1"},
                {"splits": ["task2"], "id": "task2"},
                {"splits": ["task2"], "id": "task3"},
                {"splits": ["task2"], "id": "task4"},
                {"splits": ["task2", "task5"], "id": "task5"},
                {"splits": ["task2", "task5"], "id": "task6"},
                {"splits": ["task2", "task5"], "id": "task7"},
                {"splits": ["task2", "task5", "task8"], "id": "task8"},
                {"splits": ["task2", "task5", "task8", "task9"], "id": "task9"},
                {"splits": ["task2", "task14", "task17"], "id": "task18"},
                {"id": "init"},
                {"splits": ["task2", "task14", "task17"], "id": "task19"},
                {
                    "splits": [
                        "task2",
                        "task5",
                        "task8",
                        "task9",
                        "task11",
                        "task14",
                        "task17",
                        "notify",
                    ],
                    "id": "notify",
                },
                {"splits": ["task2", "task5", "task8", "task9", "task11"], "id": "task12"},
                {"splits": ["task2", "task5", "task8", "task9", "task11"], "id": "task13"},
                {"splits": ["task2", "task5", "task8", "task9"], "id": "task10"},
                {"splits": ["task2", "task5", "task8", "task9", "task11"], "id": "task11"},
                {"splits": ["task2", "task14"], "id": "task16"},
                {
                    "splits": ["task2", "task5", "task8", "task9", "task11", "task14", "task17"],
                    "id": "task17",
                },
                {
                    "splits": ["task2", "task5", "task8", "task9", "task11", "task14"],
                    "id": "task14",
                },
                {
                    "splits": ["task2", "task5", "task8", "task9", "task11", "task14"],
                    "id": "task15",
                },
            ],
            "adjacency": [
                [{"ref": 0, "id": "task2", "key": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"ref": 2, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                    {
                        "ref": 0,
                        "id": "task3",
                        "key": 0,
                        "criteria": ["<% succeeded() and ctx().fork2 %>"],
                    },
                    {
                        "ref": 1,
                        "id": "task5",
                        "key": 0,
                        "criteria": ["<% succeeded() and not ctx().fork2 %>"],
                    },
                ],
                [{"ref": 0, "id": "task4", "key": 0, "criteria": ["<% succeeded() %>"]}],
                [{"ref": 0, "id": "task5", "key": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {
                        "ref": 0,
                        "id": "task6",
                        "key": 0,
                        "criteria": ["<% succeeded() and ctx().fork3 %>"],
                    },
                    {
                        "ref": 1,
                        "id": "task7",
                        "key": 0,
                        "criteria": ["<% succeeded() and not ctx().fork3 %>"],
                    },
                ],
                [
                    {"ref": 2, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                    {
                        "ref": 0,
                        "id": "task8",
                        "key": 0,
                        "criteria": ["<% succeeded() and ctx().fork4 %>"],
                    },
                    {
                        "ref": 1,
                        "id": "task9",
                        "key": 0,
                        "criteria": ["<% succeeded() and not ctx().fork4 %>"],
                    },
                ],
                [
                    {"ref": 2, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                    {
                        "ref": 0,
                        "id": "task8",
                        "key": 0,
                        "criteria": ["<% succeeded() and ctx().fork4 %>"],
                    },
                    {
                        "ref": 1,
                        "id": "task9",
                        "key": 0,
                        "criteria": ["<% succeeded() and not ctx().fork4 %>"],
                    },
                ],
                [{"ref": 0, "id": "task9", "key": 0, "criteria": []}],
                [
                    {"ref": 0, "id": "task10", "key": 0, "criteria": ["<% not ctx().fork5 %>"]},
                    {"ref": 1, "id": "task11", "key": 0, "criteria": ["<% ctx().fork5 %>"]},
                ],
                [
                    {
                        "ref": 1,
                        "id": "task18",
                        "key": 0,
                        "criteria": ["<% succeeded() and not result() %>"],
                    },
                    {
                        "ref": 0,
                        "id": "task19",
                        "key": 0,
                        "criteria": ["<% succeeded() and result() %>"],
                    },
                ],
                [
                    {"ref": 0, "id": "notify", "key": 0, "criteria": ["<% ctx().fork1 %>"]},
                    {"ref": 1, "id": "notify", "key": 1, "criteria": ["<% not ctx().fork1 %>"]},
                    {"ref": 0, "id": "task1", "key": 0, "criteria": ["<% ctx().fork1 %>"]},
                    {"ref": 1, "id": "task2", "key": 0, "criteria": ["<% not ctx().fork1 %>"]},
                ],
                [{"ref": 0, "id": "notify", "key": 0, "criteria": []}],
                [],
                [
                    {"ref": 0, "id": "task13", "key": 0, "criteria": ["<% succeeded() %>"]},
                    {"ref": 1, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                ],
                [
                    {"ref": 0, "id": "task14", "key": 0, "criteria": ["<% succeeded() %>"]},
                    {"ref": 1, "id": "task14", "key": 1, "criteria": ["<% failed()  %>"]},
                ],
                [{"ref": 0, "id": "task11", "key": 0, "criteria": []}],
                [
                    {"ref": 0, "id": "task12", "key": 0, "criteria": ["<% succeeded() %>"]},
                    {"ref": 1, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                ],
                [
                    {
                        "ref": 1,
                        "id": "task16",
                        "key": 0,
                        "criteria": ["<% succeeded() and not result() %>"],
                    },
                    {
                        "ref": 0,
                        "id": "task17",
                        "key": 0,
                        "criteria": ["<% succeeded() and result() %>"],
                    },
                ],
                [
                    {"ref": 1, "id": "notify", "key": 0, "criteria": ["<% not ctx().fork8 %>"]},
                    {"ref": 0, "id": "task18", "key": 0, "criteria": ["<% ctx().fork8 %>"]},
                ],
                [
                    {"ref": 0, "id": "task15", "key": 0, "criteria": ["<% not ctx().fork6 %>"]},
                    {"ref": 1, "id": "task16", "key": 0, "criteria": ["<% ctx().fork6 %>"]},
                ],
                [
                    {"ref": 0, "id": "task16", "key": 0, "criteria": ["<% ctx().fork7 %>"]},
                    {"ref": 1, "id": "task17", "key": 0, "criteria": ["<% not ctx().fork7 %>"]},
                ],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": [],
            "nodes": [
                {"id": "task1"},
                {"splits": ["task2"], "id": "task2"},
                {"splits": ["task2"], "id": "task3"},
                {"splits": ["task2"], "id": "task4"},
                {"splits": ["task2", "task5"], "id": "task5"},
                {"splits": ["task2", "task5"], "id": "task6"},
                {"splits": ["task2", "task5"], "id": "task7"},
                {"splits": ["task2", "task5", "task8"], "id": "task8"},
                {"splits": ["task2", "task5", "task8", "task9"], "id": "task9"},
                {"splits": ["task2", "task14", "task17"], "id": "task18"},
                {"id": "init"},
                {"splits": ["task2", "task14", "task17"], "id": "task19"},
                {
                    "splits": [
                        "task2",
                        "task5",
                        "task8",
                        "task9",
                        "task11",
                        "task14",
                        "task17",
                        "notify",
                    ],
                    "id": "notify",
                },
                {"splits": ["task2", "task5", "task8", "task9", "task11"], "id": "task12"},
                {"splits": ["task2", "task5", "task8", "task9", "task11"], "id": "task13"},
                {"splits": ["task2", "task5", "task8", "task9"], "id": "task10"},
                {"splits": ["task2", "task5", "task8", "task9", "task11"], "id": "task11"},
                {"splits": ["task2", "task14"], "id": "task16"},
                {
                    "splits": ["task2", "task5", "task8", "task9", "task11", "task14", "task17"],
                    "id": "task17",
                },
                {
                    "splits": ["task2", "task5", "task8", "task9", "task11", "task14"],
                    "id": "task14",
                },
                {
                    "splits": ["task2", "task5", "task8", "task9", "task11", "task14"],
                    "id": "task15",
                },
            ],
            "adjacency": [
                [{"ref": 0, "id": "task2", "key": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {"ref": 2, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                    {
                        "ref": 0,
                        "id": "task3",
                        "key": 0,
                        "criteria": ["<% succeeded() and ctx().fork2 %>"],
                    },
                    {
                        "ref": 1,
                        "id": "task5",
                        "key": 0,
                        "criteria": ["<% succeeded() and not ctx().fork2 %>"],
                    },
                ],
                [{"ref": 0, "id": "task4", "key": 0, "criteria": ["<% succeeded() %>"]}],
                [{"ref": 0, "id": "task5", "key": 0, "criteria": ["<% succeeded() %>"]}],
                [
                    {
                        "ref": 0,
                        "id": "task6",
                        "key": 0,
                        "criteria": ["<% succeeded() and ctx().fork3 %>"],
                    },
                    {
                        "ref": 1,
                        "id": "task7",
                        "key": 0,
                        "criteria": ["<% succeeded() and not ctx().fork3 %>"],
                    },
                ],
                [
                    {"ref": 2, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                    {
                        "ref": 0,
                        "id": "task8",
                        "key": 0,
                        "criteria": ["<% succeeded() and ctx().fork4 %>"],
                    },
                    {
                        "ref": 1,
                        "id": "task9",
                        "key": 0,
                        "criteria": ["<% succeeded() and not ctx().fork4 %>"],
                    },
                ],
                [
                    {"ref": 2, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                    {
                        "ref": 0,
                        "id": "task8",
                        "key": 0,
                        "criteria": ["<% succeeded() and ctx().fork4 %>"],
                    },
                    {
                        "ref": 1,
                        "id": "task9",
                        "key": 0,
                        "criteria": ["<% succeeded() and not ctx().fork4 %>"],
                    },
                ],
                [{"ref": 0, "id": "task9", "key": 0, "criteria": []}],
                [
                    {"ref": 0, "id": "task10", "key": 0, "criteria": ["<% not ctx().fork5 %>"]},
                    {"ref": 1, "id": "task11", "key": 0, "criteria": ["<% ctx().fork5 %>"]},
                ],
                [
                    {
                        "ref": 1,
                        "id": "task18",
                        "key": 0,
                        "criteria": ["<% succeeded() and not result() %>"],
                    },
                    {
                        "ref": 0,
                        "id": "task19",
                        "key": 0,
                        "criteria": ["<% succeeded() and result() %>"],
                    },
                ],
                [
                    {"ref": 0, "id": "notify", "key": 0, "criteria": ["<% ctx().fork1 %>"]},
                    {"ref": 1, "id": "notify", "key": 1, "criteria": ["<% not ctx().fork1 %>"]},
                    {"ref": 0, "id": "task1", "key": 0, "criteria": ["<% ctx().fork1 %>"]},
                    {"ref": 1, "id": "task2", "key": 0, "criteria": ["<% not ctx().fork1 %>"]},
                ],
                [{"ref": 0, "id": "notify", "key": 0, "criteria": []}],
                [],
                [
                    {"ref": 0, "id": "task13", "key": 0, "criteria": ["<% succeeded() %>"]},
                    {"ref": 1, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                ],
                [
                    {"ref": 0, "id": "task14", "key": 0, "criteria": ["<% succeeded() %>"]},
                    {"ref": 1, "id": "task14", "key": 1, "criteria": ["<% failed()  %>"]},
                ],
                [{"ref": 0, "id": "task11", "key": 0, "criteria": []}],
                [
                    {"ref": 0, "id": "task12", "key": 0, "criteria": ["<% succeeded() %>"]},
                    {"ref": 1, "id": "task14", "key": 0, "criteria": ["<% failed() %>"]},
                ],
                [
                    {
                        "ref": 1,
                        "id": "task16",
                        "key": 0,
                        "criteria": ["<% succeeded() and not result() %>"],
                    },
                    {
                        "ref": 0,
                        "id": "task17",
                        "key": 0,
                        "criteria": ["<% succeeded() and result() %>"],
                    },
                ],
                [
                    {"ref": 1, "id": "notify", "key": 0, "criteria": ["<% not ctx().fork8 %>"]},
                    {"ref": 0, "id": "task18", "key": 0, "criteria": ["<% ctx().fork8 %>"]},
                ],
                [
                    {"ref": 0, "id": "task15", "key": 0, "criteria": ["<% not ctx().fork6 %>"]},
                    {"ref": 1, "id": "task16", "key": 0, "criteria": ["<% ctx().fork6 %>"]},
                ],
                [
                    {"ref": 0, "id": "task16", "key": 0, "criteria": ["<% ctx().fork7 %>"]},
                    {"ref": 1, "id": "task17", "key": 0, "criteria": ["<% not ctx().fork7 %>"]},
                ],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.expressions import base as expr_base
from orquesta.tests.unit import base as test_base


class YAQLFacadeVariableExtractionTest(test_base.ExpressionFacadeEvaluatorTest):
    def test_empty_extraction(self):
        expr = (
            "<% just_text and $not_a_var and "
            "notctx(foo) and notctx(\"bar\") and notctx('fu') "
            'ctx("foo\') and ctx(\'foo") and ctx(foo") and '
            "ctx(\"foo) and ctx(foo') and ctx('foo) and "
            "ctx(-foo) and ctx(\"-bar\") and ctx('-fu') and "
            "ctx(foo.bar) and ctx(\"foo.bar\") and ctx('foo.bar') and "
            "ctx(foo()) and ctx(\"foo()\") and ctx('foo()') %>"
        )

        self.assertListEqual([], expr_base.extract_vars(expr))

    def test_single_var_extraction(self):
        expr = "<% ctx(foobar) %>"

        expected_vars = [("yaql", expr, "foobar")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_single_dotted_var_extraction(self):
        expr = "<% ctx(foo).bar %>"

        expected_vars = [("yaql", expr, "foo")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_single_indexing_var_extraction(self):
        expr = "<% ctx(foo)[0] %>"

        expected_vars = [("yaql", expr, "foo")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_single_functional_var_extraction(self):
        expr = "<% ctx(foo).get(bar) %>"

        expected_vars = [("yaql", expr, "foo")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_multiple_vars_extraction(self):
        expr = "<%ctx(fubar) ctx(foobar) ctx(foo).get(bar) ctx(fu).bar ctx(foobaz).bar[0] %>"

        expected_vars = [
            ("yaql", expr, "foo"),
            ("yaql", expr, "foobar"),
            ("yaql", expr, "foobaz"),
            ("yaql", expr, "fu"),
            ("yaql", expr, "fubar"),
        ]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_multiple_interleaved_vars_extraction(self):
        expr = "<% Why the ctx(foobar) are you so ctx(fu).bar serious? %>"

        expected_vars = [("yaql", expr, "foobar"), ("yaql", expr, "fu")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_vars_extraction_from_list(self):
        expr = [
            "<% abc %>",
            "<% Why the ctx(foobar) are you so ctx(fu).bar serious? %>",
            "All your test_base are belong to us.",
            {"<% ctx(x) %>": 123, "k2": "<% ctx(y) %>", "k3": ["<% ctx(z) %>"]},
        ]

        expected_vars = [
            ("yaql", expr[1], "foobar"),
            ("yaql", expr[1], "fu"),
            ("yaql", "<% ctx(x) %>", "x"),
            ("yaql", "<% ctx(y) %>", "y"),
            ("yaql", "<% ctx(z) %>", "z"),
        ]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_vars_extraction_from_dict(self):
        expr = {
            "k1": "<% abc %>",
            "k2": "<% Why the ctx(foobar) are you so ctx(fu).bar serious? %>",
            "k3": ["<% ctx(z) %>"],
            "k4": {"k5": "<% ctx(y) %>"},
            "<% ctx(x) %>": 123,
        }

        expected_vars = [
            ("yaql", expr["k2"], "foobar"),
            ("yaql", expr["k2"], "fu"),
            ("yaql", "<% ctx(x) %>", "x"),
            ("yaql", "<% ctx(y) %>", "y"),
            ("yaql", "<% ctx(z) %>", "z"),
        ]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_ignore_ctx_dict_funcs(self):
        expr = '<%ctx().keys() and ctx().values() and ctx().set("b", 3) %>'

        expected_vars = []

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_ignore_ctx_get_func_calls(self):
        expr = (
            "<%ctx().get(foo) and ctx().get(bar) and ctx().get(\"fu\") and ctx().get('baz') and "
            'ctx().get(foo, "bar") and ctx().get("fu", "bar") and ctx().get(\'bar\', \'foo\') and '
            'ctx().get("foo\') and ctx().get(\'foo") and ctx().get("foo) and ctx().get(foo") %>'
        )

        expected_vars = []

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import six
import yaml

from orquesta.specs import loader as spec_loader


LOG = logging.getLogger(__name__)


def instantiate(spec_type, definition):
    if not definition:
        raise ValueError("Workflow definition is empty.")

    if isinstance(definition, six.string_types):
        definition = yaml.safe_load(definition)

    if not isinstance(definition, dict):
        raise ValueError("Unable to convert workflow definition into dict.")

    spec_module = spec_loader.get_spec_module(spec_type)

    version = definition.pop("version", None)

    if not version:
        raise ValueError("Version of the workflow definition is not provided.")

    spec_version = spec_module.VERSION

    if str(version) != str(spec_version):
        raise ValueError('Workflow definition is not the supported version "%s".', spec_version)

    if not definition.keys():
        raise ValueError("Workflow definition contains no workflow.")

    return spec_module.instantiate(definition)


def deserialize(data):
    spec_type = data.get("catalog")
    spec_module = spec_loader.get_spec_module(spec_type)

    return spec_module.deserialize(data)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.specs import base as spec_base
from orquesta.specs import types as spec_types


class MockBaseSpec(spec_base.Spec):
    _catalog = "test"
    _version = "1.0"


class MockBaseMappingSpec(spec_base.MappingSpec):
    _catalog = "test"
    _version = "1.0"


class MockBaseSequenceSpec(spec_base.SequenceSpec):
    _catalog = "test"
    _version = "1.0"


class MockLeafSpec(MockBaseSpec):
    _schema = {
        "type": "object",
        "properties": {"attr1": spec_types.NONEMPTY_STRING, "attr2": spec_types.NONEMPTY_STRING},
        "required": ["attr1"],
        "additionalProperties": False,
    }

    _context_evaluation_sequence = ["attr1", "attr2"]


class MockJointSpec(MockBaseSpec):
    _schema = {
        "type": "object",
        "properties": {"attr1": MockLeafSpec},
        "required": ["attr1"],
        "additionalProperties": False,
    }

    _context_evaluation_sequence = ["attr1"]


class MockMappingSpec(MockBaseMappingSpec):
    _schema = {"type": "object", "patternProperties": {r"^\w+$": MockJointSpec}}


class MockSequenceSpec(MockBaseSequenceSpec):
    _schema = {"type": "array", "items": MockJointSpec}


class MockSpec(MockBaseSpec):
    _schema = {
        "type": "object",
        "properties": {
            "inputs": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
            "vars": spec_types.NONEMPTY_DICT,
            "attr1": spec_types.NONEMPTY_STRING,
            "attr1-1": spec_types.NONEMPTY_STRING,
            "attr1_2": spec_types.NONEMPTY_STRING,
            "attr2": spec_types.NONEMPTY_DICT,
            "attr3": spec_types.UNIQUE_STRING_LIST,
            "attr4": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
            "attr5": MockJointSpec,
            "attr6": MockMappingSpec,
            "attr7": MockSequenceSpec,
        },
        "required": ["attr1"],
        "additionalProperties": False,
    }

    _context_evaluation_sequence = [
        "inputs",
        "vars",
        "attr1",
        "attr1-1",
        "attr1_2",
        "attr2",
        "attr3",
        "attr4",
        "attr5",
        "attr6",
        "attr7",
    ]

    _context_inputs = ["inputs", "vars"]



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import statuses

from orquesta.tests.unit.conducting.native import base


class SequentialWorkflowStatusTest(base.OrchestraWorkflowConductorTest):
    def __init__(self, *args, **kwargs):
        super(SequentialWorkflowStatusTest, self).__init__(*args, **kwargs)
        self.wf_name = "sequential"

    def assert_workflow_status(self, mock_flow_entries, expected_wf_statuses, conductor=None):
        return super(SequentialWorkflowStatusTest, self).assert_workflow_status(
            self.wf_name, mock_flow_entries, expected_wf_statuses, conductor=conductor
        )

    def test_success(self):
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.SUCCEEDED},
            {"id": "task3", "name": "task3", "status": statuses.RUNNING},
            {"id": "task3", "name": "task3", "status": statuses.SUCCEEDED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.SUCCEEDED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_failure(self):
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.FAILED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.FAILED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_pending(self):
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.PENDING},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
        ]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Resume the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.RESUMING)

        mock_flow_entries = [
            {"id": "task2", "name": "task2", "status": statuses.SUCCEEDED},
            {"id": "task3", "name": "task3", "status": statuses.RUNNING},
            {"id": "task3", "name": "task3", "status": statuses.SUCCEEDED},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.SUCCEEDED]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_workflow_pausing_while_task_running(self):
        # Run workflow until task2 is running.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.RUNNING]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Pause the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.PAUSING)

        mock_flow_entries = [{"id": "task2", "name": "task2", "status": statuses.SUCCEEDED}]

        expected_wf_statuses = [statuses.PAUSED]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_workflow_pausing_and_resuming_while_task_running(self):
        # Run workflow until task2 is running.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.RUNNING]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Pause and resume the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.PAUSING)
        conductor.request_workflow_status(statuses.RESUMING)

        mock_flow_entries = [{"id": "task2", "name": "task2", "status": statuses.SUCCEEDED}]

        expected_wf_statuses = [statuses.RUNNING]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_workflow_pausing_then_task_abended(self):
        # Run workflow until task2 is running.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.RUNNING]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Pause the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.PAUSING)

        mock_flow_entries = [{"id": "task2", "name": "task2", "status": statuses.FAILED}]

        expected_wf_statuses = [statuses.FAILED]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_workflow_pausing_then_task_canceled(self):
        # Run workflow until task2 is running.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.RUNNING]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Pause the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.PAUSING)

        mock_flow_entries = [{"id": "task2", "name": "task2", "status": statuses.CANCELED}]

        expected_wf_statuses = [statuses.CANCELED]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_workflow_pausing_then_task_pending(self):
        # Run workflow until task2 is running.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.RUNNING]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Pause the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.PAUSING)

        mock_flow_entries = [{"id": "task2", "name": "task2", "status": statuses.PENDING}]

        expected_wf_statuses = [statuses.PAUSED]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_workflow_resuming(self):
        # Run workflow until task2 is running.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.RUNNING]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Pause the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.PAUSING)

        mock_flow_entries = [{"id": "task2", "name": "task2", "status": statuses.SUCCEEDED}]

        expected_wf_statuses = [statuses.PAUSED]

        # Assert the remaining statuses using the previous conductor.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor)

        # Resume the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.RESUMING)

        mock_flow_entries = [
            {"id": "task3", "name": "task3", "status": statuses.RUNNING},
            {"id": "task3", "name": "task3", "status": statuses.SUCCEEDED},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.SUCCEEDED]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_workflow_resuming_then_task_abended(self):
        # Run workflow until task2 is running.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.RUNNING]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Pause the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.PAUSING)

        mock_flow_entries = [{"id": "task2", "name": "task2", "status": statuses.SUCCEEDED}]

        expected_wf_statuses = [statuses.PAUSED]

        # Assert the remaining statuses using the previous conductor.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor)

        # Resume the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.RESUMING)

        mock_flow_entries = [
            {"id": "task3", "name": "task3", "status": statuses.RUNNING},
            {"id": "task3", "name": "task3", "status": statuses.FAILED},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.FAILED]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_workflow_resuming_then_task_canceled(self):
        # Run workflow until task2 is running.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.RUNNING]

        # Assert statuses and then save the conductor for later.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

        # Pause the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.PAUSING)

        mock_flow_entries = [{"id": "task2", "name": "task2", "status": statuses.SUCCEEDED}]

        expected_wf_statuses = [statuses.PAUSED]

        # Assert the remaining statuses using the previous conductor.
        conductor = self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor)

        # Resume the workflow and assert the remaining statuses.
        conductor.request_workflow_status(statuses.RESUMING)

        mock_flow_entries = [
            {"id": "task3", "name": "task3", "status": statuses.RUNNING},
            {"id": "task3", "name": "task3", "status": statuses.CANCELED},
        ]

        expected_wf_statuses = [statuses.RUNNING, statuses.CANCELED]

        # Assert the remaining statuses using the previous conductor.
        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses, conductor=conductor)

    def test_task_pause(self):
        # Test use case where a task is paused.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_task_pausing_then_canceled(self):
        # Test use case where a task is pausing and then canceled.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSING},
            {"id": "task2", "name": "task2", "status": statuses.CANCELING},
            {"id": "task2", "name": "task2", "status": statuses.CANCELED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.CANCELING,
            statuses.CANCELED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_task_pausing_then_abended(self):
        # Test use case where a task is pausing and then failed.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSING},
            {"id": "task2", "name": "task2", "status": statuses.FAILED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.FAILED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_task_resume(self):
        # Test use case where a task is paused and resumed.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSED},
            {"id": "task2", "name": "task2", "status": statuses.RESUMING},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.SUCCEEDED},
            {"id": "task3", "name": "task3", "status": statuses.RUNNING},
            {"id": "task3", "name": "task3", "status": statuses.SUCCEEDED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.SUCCEEDED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_task_resuming_then_canceled(self):
        # Test use case where a task is paused and resumed.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSED},
            {"id": "task2", "name": "task2", "status": statuses.RESUMING},
            {"id": "task2", "name": "task2", "status": statuses.CANCELING},
            {"id": "task2", "name": "task2", "status": statuses.CANCELED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.PAUSED,
            statuses.RUNNING,
            statuses.CANCELING,
            statuses.CANCELED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_task_cancel(self):
        # Test use case where a task is canceled.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.CANCELING},
            {"id": "task2", "name": "task2", "status": statuses.CANCELED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.CANCELING,
            statuses.CANCELED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_task_canceling_then_paused(self):
        # Test use case where a task is canceling and then paused.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.CANCELING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSING},
            {"id": "task2", "name": "task2", "status": statuses.PAUSED},
            {"id": "task2", "name": "task2", "status": statuses.CANCELED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.CANCELING,
            statuses.CANCELING,
            statuses.CANCELED,
            statuses.CANCELED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)

    def test_task_canceling_then_abended(self):
        # Test use case where a task is pausing and then failed.
        mock_flow_entries = [
            {"id": "task1", "name": "task1", "status": statuses.RUNNING},
            {"id": "task1", "name": "task1", "status": statuses.SUCCEEDED},
            {"id": "task2", "name": "task2", "status": statuses.RUNNING},
            {"id": "task2", "name": "task2", "status": statuses.CANCELING},
            {"id": "task2", "name": "task2", "status": statuses.FAILED},
        ]

        expected_wf_statuses = [
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.RUNNING,
            statuses.CANCELING,
            statuses.CANCELED,
        ]

        self.assert_workflow_status(mock_flow_entries, expected_wf_statuses)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta import exceptions as exc
from orquesta.expressions.functions import workflow as funcs


class ItemFunctionTest(unittest.TestCase):
    def test_missing_current_item(self):
        self.assertRaises(exc.ExpressionEvaluationException, funcs.item_, {})

    def test_item_is_null(self):
        context = {"__current_item": None}
        self.assertIsNone(funcs.item_(context))

    def test_item(self):
        context = {"__current_item": "foobar"}
        self.assertEqual(funcs.item_(context), "foobar")

    def test_item_is_dict(self):
        context = {"__current_item": {"foo": "bar"}}
        self.assertDictEqual(funcs.item_(context), {"foo": "bar"})

    def test_item_is_not_dict(self):
        context = {"__current_item": "foobar"}
        self.assertRaises(exc.ExpressionEvaluationException, funcs.item_, context, key="foo")

    def test_item_key(self):
        context = {"__current_item": {"foo": "bar"}}
        self.assertEqual(funcs.item_(context, key="foo"), "bar")

    def test_item_bad_key(self):
        context = {"__current_item": {"foo": "bar"}}
        self.assertRaises(exc.ExpressionEvaluationException, funcs.item_, context, key="bar")



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.conducting.mistral import base


class SplitWorkflowConductorTest(base.MistralWorkflowConductorTest):
    def test_split(self):
        wf_name = "split"

        expected_routes = [
            [],  # default from start
            ["task2__t0"],  # task1 -> task2 -> task4
            ["task3__t0"],  # task1 -> task3 -> task4
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 0),
            ("task4", 1),
            ("task4", 2),
            ("task5", 1),
            ("task5", 2),
            ("task6", 1),
            ("task6", 2),
            ("task7", 1),
            ("task7", 2),
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq, expected_routes)

    def test_splits(self):
        wf_name = "splits"

        expected_routes = [
            [],  # default from start
            ["task1__t0"],  # task1 -> task8
            ["task2__t0"],  # task1 -> task2 -> task4
            ["task3__t0"],  # task1 -> task3 -> task4
            ["task2__t0", "task7__t0"],  # task1 -> task2 -> task4 -> task5/6 -> task7 -> task8
            ["task3__t0", "task7__t0"],  # task1 -> task3 -> task4 -> task5/6 -> task7 -> task8
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 0),
            ("task8", 1),
            ("task4", 2),
            ("task4", 3),
            ("task5", 2),
            ("task5", 3),
            ("task6", 2),
            ("task6", 3),
            ("task7", 2),
            ("task7", 3),
            ("task8", 4),
            ("task8", 5),
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq, expected_routes)

    def test_nested_splits(self):
        wf_name = "splits-nested"

        expected_routes = [
            [],  # default from start
            ["task2__t0"],  # task1 -> task2 -> task4
            ["task3__t0"],  # task1 -> task3 -> task4
            ["task2__t0", "task5__t0"],  # task1 -> task2 -> task4 -> task5 -> task7
            ["task3__t0", "task5__t0"],  # task1 -> task3 -> task4 -> task5 -> task7
            ["task2__t0", "task6__t0"],  # task1 -> task2 -> task4 -> task6 -> task7
            ["task3__t0", "task6__t0"],  # task1 -> task3 -> task4 -> task6 -> task7
        ]

        expected_task_seq = [
            ("task1", 0),
            ("task2", 0),
            ("task3", 0),
            ("task4", 1),
            ("task4", 2),
            ("task5", 1),
            ("task5", 2),
            ("task6", 1),
            ("task6", 2),
            ("task7", 3),
            ("task7", 4),
            ("task7", 5),
            ("task7", 6),
            ("task8", 3),
            ("task8", 4),
            ("task8", 5),
            ("task8", 6),
            ("task9", 3),
            ("task9", 4),
            ("task9", 5),
            ("task9", 6),
            ("task10", 3),
            ("task10", 4),
            ("task10", 5),
            ("task10", 6),
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq, expected_routes)
